/*
 * Zorin Tray: Extension to manage legacy tray icons in Zorin OS.
 *
 * Original work: Copyright (C) phocean <jc@phocean.net>
 *                Copyright (C) ag
 *                Copyright (C) Mjnaderi
 * Modified work: Copyright (C) 2016-2017 Zorin OS Technologies Ltd.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * Credits:
 * This file is based on code from the TopIcons-Plus extension by phocean
 * (as at commit 41f1a9b5dcf4b0c92ffce378b3e772253be479e5).
 */

const Shell = imports.gi.Shell;
const St = imports.gi.St;
const Main = imports.ui.main;
const GLib = imports.gi.GLib;
const Lang = imports.lang;
const System = imports.system;
const PanelMenu = imports.ui.panelMenu;
const ExtensionUtils = imports.misc.extensionUtils;

const ICON_SIZE = 24;

let tray = null;
let trayIconImplementations = null;
let trayAddedId = 0;
let trayRemovedId = 0;
let icons = [];
let iconsBoxLayout = null;
let iconsContainer = null;
let panelChildSignals = {};
let blacklist = [["skype","SkypeNotification@chrisss404.gmail.com"]]; // blacklist: array of [uuid, wmClass (icon application name)] pairs

function init() { }

function enable() {
    tray = Main.legacyTray;

    if (tray)
        GLib.idle_add(GLib.PRIORITY_LOW, moveToTop);
    else
        GLib.idle_add(GLib.PRIORITY_LOW, createTray);

    connectPanelChildSignals();
}

function disable() {

    if (Main.legacyTray)
        moveToTray();
    else
        destroyTray();

    disconnectPanelChildSignals();
}

function onTrayIconAdded(o, icon, role, delay=1000) {

    let iconContainer = new St.Button({child: icon, visible: false});

    icon.connect("destroy", function() {
        iconContainer.destroy();
    });

    iconContainer.connect('button-release-event', function(actor, event) {
        icon.click(event);
    });

    iconsBoxLayout.insert_child_at_index(iconContainer, 0);

    if (checkApp(icon))
        GLib.timeout_add(GLib.PRIORITY_DEFAULT, delay, Lang.bind(this, function(){
            iconContainer.visible = true;
            iconsContainer.actor.visible = true;
            return GLib.SOURCE_REMOVE;
        }));
    else
         GLib.timeout_add(GLib.PRIORITY_DEFAULT, delay, Lang.bind(this, function(){
            iconContainer.visible = false;
            return GLib.SOURCE_REMOVE;
        }));

    setIcon(icon);
    icons.push(icon);
}

function onTrayIconRemoved(o, icon) {

    if (icons.indexOf(icon) == -1) {
      return;
    }

    let parent = icon.get_parent();
    if (parent)
         parent.destroy();
    icon.destroy();
    icons.splice(icons.indexOf(icon), 1);

    if (icons.length === 0)
        iconsContainer.actor.visible = false;
    
}

function onPanelChange(actor, child) {
    if (!iconsBoxLayout || iconsBoxLayout.get_parent() === child)
        return;

    // refresh position on panel left/center/right
    // box add/remove child event
    placeTray();
}

function connectPanelChildSignals() {
    panelChildSignals = {
        left: {
            add: Main.panel._leftBox.connect('actor_added', Lang.bind(this, onPanelChange)),
            del: Main.panel._leftBox.connect('actor_removed', Lang.bind(this, onPanelChange))
        },
        center: {
            add: Main.panel._centerBox.connect('actor_added', Lang.bind(this, onPanelChange)),
            del: Main.panel._centerBox.connect('actor_removed', Lang.bind(this, onPanelChange))
        },
        right: {
            add: Main.panel._rightBox.connect('actor_added', Lang.bind(this, onPanelChange)),
            del: Main.panel._rightBox.connect('actor_removed', Lang.bind(this, onPanelChange))
        }
    }
}

function disconnectPanelChildSignals() {
    Main.panel._leftBox.disconnect(panelChildSignals.left.add);
    Main.panel._leftBox.disconnect(panelChildSignals.left.del);
    Main.panel._centerBox.disconnect(panelChildSignals.center.add);
    Main.panel._centerBox.disconnect(panelChildSignals.center.del);
    Main.panel._rightBox.disconnect(panelChildSignals.right.add);
    Main.panel._rightBox.disconnect(panelChildSignals.right.del);
}

function createIconsContainer() {
    // Create box layout for icon containers
    iconsBoxLayout = new St.BoxLayout();
    setSpacing();

    // An empty ButtonBox will still display padding,therefore create it without visibility.
    iconsContainer = new PanelMenu.ButtonBox({visible: false});
    iconsContainer.actor.add_actor(iconsBoxLayout);
}

function createTray() {
    createIconsContainer();

    tray = new Shell.TrayManager();
    tray.connect('tray-icon-added', onTrayIconAdded);
    tray.connect('tray-icon-removed', onTrayIconRemoved);
    tray.manage_screen(global.screen, Main.panel.actor);
    placeTray();
}

function destroyTray() {
    iconsContainer.actor.destroy();
    iconsContainer = null;
    iconsBoxLayout = null;
    icons = [];

    tray = null;
    System.gc(); // force finalizing tray to unmanage screen
}

function moveToTop() {

    // Replace signal handlers
    if (tray._trayIconAddedId)
        tray._trayManager.disconnect(tray._trayIconAddedId);
    if (tray._trayIconRemovedId)
        tray._trayManager.disconnect(tray._trayIconRemovedId);
    trayAddedId = tray._trayManager.connect('tray-icon-added', onTrayIconAdded);
    trayRemovedId = tray._trayManager.connect('tray-icon-removed', onTrayIconRemoved);

    createIconsContainer();
    placeTray();

    // Move each tray icon to the top
    let length = tray._iconBox.get_n_children();
    for (let i = 0; i < length; i++) {
        let button = tray._iconBox.get_child_at_index(0);
        let icon = button.child;
        button.remove_actor(icon);
        button.destroy();
        // Icon already loaded, no need to delay insertion
        onTrayIconAdded(this, icon, '', 0);
    }

}

function moveToTray() {

    // Replace signal handlers
    if (trayAddedId) {
        tray._trayManager.disconnect(trayAddedId);
        trayAddedId = 0;
    }

    if (trayRemovedId) {
        tray._trayManager.disconnect(trayRemovedId);
        trayRemovedId = 0;
    }

    tray._trayIconAddedId = tray._trayManager.connect(
        'tray-icon-added', Lang.bind(tray, tray._onTrayIconAdded));
    tray._trayIconRemovedId = tray._trayManager.connect(
        'tray-icon-removed', Lang.bind(tray, tray._onTrayIconRemoved));

    // Clean and move each icon back to the Legacy Tray;
    for (let i = 0; i < icons.length; i++) {
        let icon = icons[i];
        let parent = icon.get_parent();
        if (parent) {
            parent.remove_actor(icon);
            parent.destroy();
        }
        tray._onTrayIconAdded(tray, icon);
    }

    // Clean containers
    icons = [];
    if (iconsBoxLayout) {
        iconsBoxLayout.destroy();
        iconsBoxLayout = null;
    }
    if (iconsContainer) {
        if (iconsContainer.actor) {
            iconsContainer.actor.destroy();
            iconsContainer.actor = null;
        }
        iconsContainer = null;
    }

}

function placeTray() {

    let parent = iconsContainer.actor.get_parent();
    if (parent)
        parent.remove_actor(iconsContainer.actor);
    
    Main.panel._rightBox.insert_child_at_index(iconsContainer.actor, 0);

}


function setIcon(icon) {

    let scaleFactor = St.ThemeContext.get_for_stage(global.stage).scale_factor;

    icon.reactive = true;

    icon.get_parent().set_size(ICON_SIZE * scaleFactor, ICON_SIZE * scaleFactor);
    icon.set_size(ICON_SIZE * scaleFactor, ICON_SIZE * scaleFactor);
    setSpacing();

}

function checkApp(icon) {

    // loop through the array and hide the extension if extension X is enabled and corresponding application is running
    let iconWmClass = icon.wm_class ? icon.wm_class.toLowerCase() : '';
    for (let [wmClass, uuid] of blacklist) {
        if (ExtensionUtils.extensions[uuid] !== undefined &&
            ExtensionUtils.extensions[uuid].state === 1 &&
            iconWmClass === wmClass)
            return false;
        }

    return true;

}

function setSpacing() {

    iconsBoxLayout.set_style('spacing: 12px; margin_top: 2px; margin_bottom: 2px;');

}
